use crate::result::error_to_int;
use std::alloc::{alloc, alloc_zeroed, handle_alloc_error, Layout};
use std::io;
use std::mem::{align_of, forget, size_of};
use std::ops::Deref;
use std::ptr::null;
use std::slice;
use std::str;
use std::string::String as RustString;

/// The alignment to use for Inko objects.
const ALIGNMENT: usize = align_of::<usize>();

pub(crate) fn allocate(layout: Layout) -> *mut u8 {
    unsafe {
        let ptr = alloc(layout);

        if ptr.is_null() {
            handle_alloc_error(layout);
        } else {
            ptr
        }
    }
}

pub(crate) unsafe fn header_of<'a, T>(ptr: *const T) -> &'a mut Header {
    &mut *(ptr as *mut Header)
}

/// The header used by heap allocated objects.
///
/// The layout is fixed to ensure we can safely assume certain fields are at
/// certain offsets in an object, even when not knowing what type of object
/// we're dealing with.
#[repr(C)]
pub struct Header {
    /// The type of the object.
    pub instance_of: TypePointer,

    /// The number of references to the object of this header.
    ///
    /// If this count overflows the program terminates. In practise this should
    /// never happen, as one needs _a lot_ of references to achieve this.
    ///
    /// We're using a u32 here instead of a u16, as the likelihood of
    /// overflowing a u32 is very tiny, but overflowing a u16 is something that
    /// _could_ happen (i.e. a process reference shared with many other
    /// processes).
    ///
    /// For regular objects, this field is initially set to 0, while for atomic
    /// values it defaults to 1. The latter is done as atomics always use a
    /// checked decrement, so starting with 1 ensures we don't underflow this
    /// value.
    pub references: u32,
}

impl Header {
    pub(crate) fn init_atomic(&mut self, instance_of: TypePointer) {
        self.instance_of = instance_of;
        self.references = 1;
    }
}

/// A function bound to an object.
///
/// Methods don't have headers as there's no need for any, as methods aren't
/// values one can pass around in Inko.
#[repr(C)]
pub struct Method {
    /// The hash of this method, used when performing dynamic dispatch.
    pub hash: u64,

    /// A pointer to the native function that backs this method.
    pub code: extern "system" fn(),
}

/// An Inko type.
#[repr(C)]
pub struct Type {
    /// The name of the type.
    pub(crate) name: RustString,

    /// The size (in bytes) of instances of this type.
    ///
    /// This is only used for allocating and deallocating processes, as the
    /// deallocation happens asynchronously _after_ a process finishes running.
    /// For regular types the allocation and free logic is generated by the
    /// compiler.
    pub(crate) instance_size: u32,

    /// The number of method slots this type has.
    ///
    /// The actual number of methods may be less than this value.
    pub(crate) method_slots: u16,

    /// The methods of this type, as pointers to native functions.
    ///
    /// Methods are accessed frequently, and we want to do so with as little
    /// indirection and as cache-friendly as possible. For this reason we use a
    /// flexible array member, instead of a Vec.
    ///
    /// The length of this array _must_ be a power of two.
    pub methods: [Method; 0],
}

impl Type {
    pub(crate) fn alloc(
        name: RustString,
        methods: u16,
        size: u32,
    ) -> TypePointer {
        let ptr = unsafe {
            let layout = Self::layout(methods);

            // For types we zero memory out, so unused method slots are set to
            // zeroed memory, instead of random garbage.
            let ptr = alloc_zeroed(layout) as *mut Type;

            if ptr.is_null() {
                handle_alloc_error(layout);
            }

            ptr
        };
        let obj = unsafe { &mut *ptr };

        init!(obj.name => name);
        init!(obj.instance_size => size);
        init!(obj.method_slots => methods);

        TypePointer(ptr)
    }

    /// Returns a new type for a regular object.
    pub(crate) fn object(
        name: RustString,
        size: u32,
        methods: u16,
    ) -> TypePointer {
        Self::alloc(name, methods, size)
    }

    /// Returns a new type for a process.
    pub(crate) fn process(
        name: RustString,
        size: u32,
        methods: u16,
    ) -> TypePointer {
        Self::alloc(name, methods, size)
    }

    /// Returns the `Layout` for a type itself.
    pub(crate) unsafe fn layout(methods: u16) -> Layout {
        let size = size_of::<Type>() + (methods as usize * size_of::<Method>());

        Layout::from_size_align_unchecked(size, align_of::<Type>())
    }

    pub(crate) unsafe fn instance_layout(&self) -> Layout {
        Layout::from_size_align_unchecked(
            self.instance_size as usize,
            ALIGNMENT,
        )
    }
}

/// A pointer to a type.
#[repr(transparent)]
#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct TypePointer(pub(crate) *mut Type);

impl Deref for TypePointer {
    type Target = Type;

    fn deref(&self) -> &Type {
        unsafe { &*(self.0 as *const Type) }
    }
}

/// An FFI-safe primitive string.
///
/// These values are created either by the runtime or the standard library, and
/// ownership depends on the context.
///
/// If this struct's layout changes, the type `std.string.PrimitiveString` must
/// be updated accordingly.
#[repr(C)]
pub struct PrimitiveString {
    pub bytes: *const u8,
    pub size: i64,
}

impl PrimitiveString {
    pub(crate) fn empty() -> PrimitiveString {
        Self { bytes: null(), size: 0 }
    }

    pub(crate) fn error(error: io::Error) -> PrimitiveString {
        Self { bytes: error_to_int(error) as *const u8, size: -1 }
    }

    pub(crate) fn owned(value: RustString) -> PrimitiveString {
        let vec = value.into_bytes();
        let ptr = vec.as_ptr();
        let len = vec.len() as i64;

        forget(vec);
        PrimitiveString { bytes: ptr, size: len }
    }

    pub(crate) fn borrowed(value: &str) -> PrimitiveString {
        Self { bytes: value.as_ptr(), size: value.len() as i64 }
    }

    pub(crate) fn as_str<'a>(&self) -> &'a str {
        // Safety: the data being valid UTF8 is enforced when self is created.
        unsafe {
            let slice = slice::from_raw_parts(self.bytes, self.size as usize);

            str::from_utf8_unchecked(slice)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::test::OwnedType;
    use std::mem::{align_of, size_of};
    use std::ptr::addr_of;

    extern "system" fn dummy() {}

    #[test]
    fn test_header_field_offsets() {
        let header =
            Header { instance_of: TypePointer(0x7 as _), references: 42 };
        let base = addr_of!(header) as usize;

        assert_eq!(addr_of!(header.instance_of) as usize - base, 0);
        assert_eq!(addr_of!(header.references) as usize - base, 8);
    }

    #[test]
    fn test_type_field_offsets() {
        let typ = OwnedType(Type::alloc("A".to_string(), 4, 8));
        let base = (typ.0).0 as usize;

        assert_eq!(addr_of!(typ.name) as usize - base, 0);
        assert_eq!(addr_of!(typ.instance_size) as usize - base, 24);
        assert_eq!(addr_of!(typ.method_slots) as usize - base, 28);
        assert_eq!(addr_of!(typ.methods) as usize - base, 32);
    }

    #[test]
    fn test_method_field_offsets() {
        let method = Method { hash: 42, code: dummy };
        let base = addr_of!(method) as usize;

        assert_eq!(addr_of!(method.hash) as usize - base, 0);
        assert_eq!(addr_of!(method.code) as usize - base, 8);
    }

    #[test]
    fn test_type_sizes() {
        assert_eq!(size_of::<Header>(), 16);
        assert_eq!(size_of::<Method>(), 16);
        assert_eq!(size_of::<Method>(), 16);
        assert_eq!(size_of::<Type>(), 32);
    }

    #[test]
    fn test_type_alignments() {
        assert_eq!(align_of::<Header>(), ALIGNMENT);
        assert_eq!(align_of::<Method>(), ALIGNMENT);
        assert_eq!(align_of::<Type>(), ALIGNMENT);
    }

    #[test]
    fn test_type_alloc() {
        let typ = OwnedType(Type::alloc("A".to_string(), 0, 24));

        assert_eq!(typ.method_slots, 0);
        assert_eq!(typ.instance_size, 24);
    }

    #[test]
    fn test_type_object() {
        let typ = OwnedType(Type::object("A".to_string(), 24, 0));

        assert_eq!(typ.method_slots, 0);
        assert_eq!(typ.instance_size, 24);
    }

    #[test]
    fn test_type_process() {
        let typ = OwnedType(Type::process("A".to_string(), 24, 0));

        assert_eq!(typ.method_slots, 0);
        assert_eq!(typ.instance_size, 24);
    }
}
